@@ -21,25 +21,29 @@ import time
 import datetime
 import logging
 import queue
-import shutil
 import json
 
 from pathlib import Path
 from threading import Thread
-# from socketIO_client import SocketIO, SocketIONamespace
-from socketio import (
-    ClientNamespace,
-    Client as SocketIO
-)
+from socketio import ClientNamespace, Client as SocketIO
 from gevent.pywsgi import WSGIServer
-
-from . import globals as cs
+from enum import Enum
 
 from vantage6.common.docker_addons import ContainerKillListener
-from vantage6.node.docker_manager import DockerManager
+from vantage6.common.globals import VPN_CONFIG_FILE
+from vantage6.node.globals import NODE_PROXY_SERVER_HOSTNAME
 from vantage6.node.server_io import NodeClient
-from vantage6.node import proxy_server
+from vantage6.node.proxy_server import app
 from vantage6.node.util import logger_name
+from vantage6.node.docker.docker_manager import DockerManager
+from vantage6.node.docker.network_manager import IsolatedNetworkManager
+from vantage6.node.docker.vpn_manager import VPNManager
+
+
+class VPNConnectMode(Enum):
+    FIRST_TRY = 1
+    REFRESH_KEYPAIR = 2
+    REFRESH_COMPLETE = 3
 
 
 class NodeTaskNamespace(ClientNamespace):
@@ -55,12 +59,6 @@ class NodeTaskNamespace(ClientNamespace):
         super().__init__(*args, **kwargs)
         self.log = logging.getLogger(logger_name(__name__))
 
-    def on_connect(self):
-        """On connect or reconnect"""
-        self.log.info('Reconnected to the server')
-        self.node_worker_ref.sync_task_queue_with_server()
-        self.log.debug("Tasks synced again with the server...")
-
     def on_message(self, data):
         self.log.info(data)
 
@@ -99,8 +97,10 @@ class NodeTaskNamespace(ClientNamespace):
         self.log.debug("Token refreshed")
         self.node_worker_ref.connect_to_socket()
         self.log.debug("Connected to socket")
-        self.node_worker_ref.sync_task_queue_with_server()
-        self.log.debug("Tasks synced again with the server...")
+
+        # FIXME: This won't work: you're trying to access a private method!?
+        # self.node_worker_ref.__sync_task_queue_with_server()
+        # self.log.debug("Tasks synced again with the server...")
 
 
 # ------------------------------------------------------------------------------
@@ -128,7 +128,17 @@ class Node(object):
         self.log = logging.getLogger(logger_name(__name__))
 
         self.ctx = ctx
-        self.config = ctx.config
+
+        # Initialize the node. If it crashes, shut down the parts that started
+        # already
+        try:
+            self.initialize()
+        except Exception as e:
+            self.cleanup()
+            raise
+
+    def initialize(self):
+        self.config = self.ctx.config
         self.queue = queue.Queue()
         self._using_encryption = None
 
@@ -160,104 +170,37 @@ class Node(object):
         self.connect_to_socket()
 
         # Check if new tasks were posted while offline.
-        # self.log.debug("Fetching tasks that were posted while offline")
-        # self.sync_task_queue_with_server()
+        self.log.debug("Fetching tasks that were posted while offline")
+        self.__sync_task_queue_with_server()
 
-        # If we're in a 'regular' context, we'll copy the dataset to our data
-        # dir and mount it in any algorithm container that's run; bind mounts
-        # on a folder will work just fine.
-        #
-        # If we're running in dockerized mode we *cannot* bind mount a folder,
-        # because the folder is in the container and not in the host. We'll
-        # have to use a docker volume instead. This means:
-        #  1. we need to know the name of the volume so we can pass it along
-        #  2. need to have this volume mounted so we can copy files to it.
-        #
-        #  Ad 1: We'll use a default name that can be overridden by an
-        #        environment variable.
-        #  Ad 2: We'll expect `ctx.data_dir` to point to the right place. This
-        #        is OK, since ctx will be a DockerNodeContext.
-        #
-        #  This also means that the volume will have to be created & mounted
-        #  *before* this node is started, so we won't do anything with it here.
+        # setup docker isolated network manager
+        isolated_network_mgr = \
+            IsolatedNetworkManager(self.ctx.docker_network_name)
 
-        # We'll create a subfolder in the data_dir. We need this subfolder so
-        # we can easily mount it in the algorithm containers; the root folder
-        # may contain the private key, which which we don't want to share.
-        # We'll only do this if we're running outside docker, otherwise we
-        # would create '/data' on the data volume.
-        if not ctx.running_in_docker:
-            task_dir = ctx.data_dir / 'data'
-            os.makedirs(task_dir, exist_ok=True)
+        # Setup tasks dir
+        self._set_task_dir(self.ctx)
 
-        else:
-            task_dir = ctx.data_dir
+        # Setup VPN connection
+        self.vpn_manager = self.setup_vpn_connection(
+            isolated_network_mgr, self.ctx)
 
+        # setup the docker manager
         self.log.debug("Setting up the docker manager")
         self.__docker = DockerManager(
-            allowed_images=self.config.get("allowed_images"),
-            tasks_dir=task_dir,
-            isolated_network_name=f"{ctx.docker_network_name}-net",
-            node_name=ctx.name,
-            data_volume_name=ctx.docker_volume_name,
+            ctx=self.ctx,
+            isolated_network_mgr=isolated_network_mgr,
+            vpn_manager=self.vpn_manager,
+            tasks_dir=self.__tasks_dir
         )
 
-        # login to the registries
-        self.__docker.login_to_registries(
-            self.ctx.config.get("docker_registries", [])
-        )
-
-        databases = {}
-
-        # Check that the `default` database label is present. If this is
-        # not the case, older algorithms will break
-        db_labels = self.config['databases'].keys()
-        if 'default' not in db_labels:
-            self.log.error("'default' database not specified in the config!")
-            self.log.debug(f'databases in config={db_labels}')
-
-        # If we're running in a docker container, database_uri would point
-        # to a path on the *host* (since it's been read from the config
-        # file). That's no good here. Therefore, we expect the CLI to set
-        # the environment variables for us. This has the added bonus that we
-        # can override the URI from the command line as well.
-        for label in db_labels:
-            label_upper = label.upper()
-            if self.__docker.running_in_docker():
-                uri = os.environ[f'{label_upper}_DATABASE_URI']
-            else:
-                uri = self.config['databases'][label]
-
-            db_is_file = Path(uri).exists()
-            if db_is_file:
-                # We'll copy the file to the folder `data` in our task_dir.
-                self.log.info(f'Copying {uri} to {task_dir}')
-                shutil.copy(uri, task_dir)
-
-                # Since we've copied the database to the folder 'data' in the
-                # root of the volume: '/data/<database.csv>'. We'll just keep
-                # the basename (i.e. filename + ext).
-                uri = os.path.basename(uri)
-
-                self.__docker.database_is_file = True
-
-            databases[label] = {'uri': uri, 'is_file': db_is_file}
-
-        # Connect to the isolated algorithm network *only* if we're running in
-        # a docker container.
-        if ctx.running_in_docker:
-            self.__docker.connect_to_isolated_network(
-                ctx.docker_container_name,
-                aliases=[cs.NODE_PROXY_SERVER_HOSTNAME]
+        # Connect the node to the isolated algorithm network *only* if we're
+        # running in a docker container.
+        if self.ctx.running_in_docker:
+            isolated_network_mgr.connect(
+                container_name=self.ctx.docker_container_name,
+                aliases=[NODE_PROXY_SERVER_HOSTNAME]
             )
 
-        # Let's keep it safe
-        self.__docker.set_databases(databases)
-
-        # Load additional environment vars for the algorithms. This is
-        # for example usefull when a password is needed for the database
-        self.__docker.algorithm_env = self.config.get('algorithm_env', {})
-
         # Thread for sending results to the server when they come available.
         self.log.debug("Start thread for sending messages (results)")
         t = Thread(target=self.__speaking_worker, daemon=True)
@@ -279,14 +222,14 @@ class Node(object):
         """
         # supply the proxy server with a destination (the central server)
         # we might want to not use enviroment vars
-        # os.environ["SERVER_URL"] = self.server_io.host
-        # os.environ["SERVER_PORT"] = self.server_io.port
-        # os.environ["SERVER_PATH"] = self.server_io.path
+        os.environ["SERVER_URL"] = self.server_io.host
+        os.environ["SERVER_PORT"] = self.server_io.port
+        os.environ["SERVER_PATH"] = self.server_io.path
 
         if self.ctx.running_in_docker:
-            # cs.NODE_PROXY_SERVER_HOSTNAME points to the name of the proxy
+            # NODE_PROXY_SERVER_HOSTNAME points to the name of the proxy
             # when running in the isolated docker network.
-            default_proxy_host = cs.NODE_PROXY_SERVER_HOSTNAME
+            default_proxy_host = NODE_PROXY_SERVER_HOSTNAME
         else:
             # If we're running non-dockerized, assume that the proxy is
             # accessible from within the docker algorithm container on
@@ -302,13 +245,13 @@ class Node(object):
 
         # 'app' is defined in vantage6.node.proxy_server
         # app.debug = True
-        proxy_server.app.config["SERVER_IO"] = self.server_io
-        proxy_server.server_url = self.server_io.base_path
+        app.config["SERVER_IO"] = self.server_io
 
+        # this is where we try to find a port for the proxyserver
         for try_number in range(5):
             self.log.info(
                 f"Starting proxyserver at '{proxy_host}:{proxy_port}'")
-            http_server = WSGIServer(('0.0.0.0', proxy_port), proxy_server.app)
+            http_server = WSGIServer(('0.0.0.0', proxy_port), app)
 
             try:
                 http_server.serve_forever()
@@ -330,7 +273,7 @@ class Node(object):
                 self.log.error('Proxyserver could not be started or crashed!')
                 self.log.error(e)
 
-    def sync_task_queue_with_server(self):
+    def __sync_task_queue_with_server(self):
         """ Get all unprocessed tasks from the server for this node."""
         assert self.server_io.cryptor, "Encrpytion has not been setup"
 
@@ -376,16 +319,28 @@ class Node(object):
 
         # Run the container. This adds the created container/task to the list
         # __docker.active_tasks
-        local_db = task['database'] if task['database'] else 'default'
-        self.__docker.run(
+        vpn_port = self.__docker.run(
             result_id=taskresult["id"],
             image=task["image"],
             docker_input=taskresult['input'],
             tmp_vol_name=vol_name,
-            token=token,
-            database=local_db
+            token=token
         )
 
+        if vpn_port:
+            # Save port of VPN client container at which it redirects traffic
+            # to the algorithm container
+            self.server_io.request(
+                f"result/{taskresult['id']}", json={"port": vpn_port},
+                method="PATCH"
+            )
+            # Save IP address of VPN container
+            node_id = self.server_io.whoami.id_
+            node_ip = self.vpn_manager.get_vpn_ip()
+            self.server_io.request(
+                f"node/{node_id}", json={"ip": node_ip}, method="PATCH"
+            )
+
     def __listening_worker(self):
         """ Listen for incoming (websocket) messages from the server.
 
@@ -400,11 +355,7 @@ class Node(object):
             # incoming messages are handled by the action_handler instance
             # which is attached when the socket connection was made. wait()
             # is blocks forever (if no time is specified).
-            try:
-                self.socketIO.wait()
-            except Exception as e:
-                self.log.critical('Listening thread exception')
-                self.log.debug(e)
+            self.socketIO.wait()
 
     def __speaking_worker(self):
         """ Sending messages to central server.
@@ -418,60 +369,54 @@ class Node(object):
         self.log.debug("Waiting for results to send to the server")
 
         while True:
-            try:
-                results = self.__docker.get_result()
+            results = self.__docker.get_result()
 
-                # notify all of a crashed container
-                if results.status_code:
-                    self.socketIO.emit(
-                        'container_failed',
-                        data={
-                            'node_id': self.server_io.whoami.id_,
-                            'status_code': results.status_code,
-                            'result_id': results.result_id,
-                            'collaboration_id': self.server_io.collaboration_id
-                        },
-                        namespace='/tasks'
-                    )
-
-                self.log.info(
-                    f"Sending result (id={results.result_id}) to the server!")
-
-                # FIXME: why are we retrieving the result *again*? Shouldn't we
-                # just store the task_id when retrieving the task the first
-                # time?
-                response = self.server_io.request(
-                    f"result/{results.result_id}")
-                task_id = response.get("task").get("id")
-
-                if not task_id:
-                    self.log.error(
-                        f"task_id of result (id={results.result_id}) "
-                        f"could not be retrieved"
-                    )
-                    return
-
-                response = self.server_io.request(f"task/{task_id}")
-                initiator_id = response.get("initiator")
-
-                if not initiator_id:
-                    self.log.error(
-                        f"Initiator id from task (id={task_id})could not be "
-                        f"retrieved"
-                    )
-
-                self.server_io.patch_results(
-                    id=results.result_id,
-                    initiator_id=initiator_id,
-                    result={
-                        'result': results.data,
-                        'log': results.logs,
-                        'finished_at': datetime.datetime.now().isoformat(),
-                    }
+            # notify all of a crashed container
+            if results.status_code:
+                self.socketIO.emit(
+                    'container_failed',
+                    data={
+                        'node_id': self.server_io.whoami.id_,
+                        'status_code': results.status_code,
+                        'result_id': results.result_id,
+                        'collaboration_id': self.server_io.collaboration_id
+                    },
+                    namespace='/tasks'
                 )
-            except Exception as e:
-                self.log.critical('Spreaker thread exception')
-                self.log.debug(e)
+
+            self.log.info(
+                f"Sending result (id={results.result_id}) to the server!")
+
+            # FIXME: why are we retrieving the result *again*? Shouldn't we
+            # just store the task_id when retrieving the task the first time?
+            response = self.server_io.request(f"result/{results.result_id}")
+            task_id = response.get("task").get("id")
+
+            if not task_id:
+                self.log.error(
+                    f"task_id of result (id={results.result_id}) "
+                    f"could not be retrieved"
+                )
+                return
+
+            response = self.server_io.request(f"task/{task_id}")
+            initiator_id = response.get("initiator")
+
+            if not initiator_id:
+                self.log.error(
+                    f"Initiator id from task (id={task_id})could not be "
+                    f"retrieved"
+                )
+
+            self.server_io.patch_results(
+                id=results.result_id,
+                initiator_id=initiator_id,
+                result={
+                    'result': results.data,
+                    'log': results.logs,
+                    'finished_at': datetime.datetime.now().isoformat(),
+                }
+            )
 
     def authenticate(self):
         """ Authenticate to the central server
@@ -482,9 +427,7 @@ class Node(object):
         api_key = self.config.get("api_key")
 
         keep_trying = True
-        i = 0
-        while keep_trying and i < 10:
-            i = i + 1
+        while keep_trying:
             try:
                 self.server_io.authenticate(api_key)
 
@@ -527,7 +470,7 @@ class Node(object):
 
         if encrypted_collaboration != encrypted_node:
             # You can't force it if it just ain't right, you know?
-            raise Exception("Expectations on encryption don't match!?")
+            raise Exception("Expectations on encryption don't match?!")
 
         if encrypted_collaboration:
             self.log.warn('Enabling encryption!')
@@ -538,6 +481,147 @@ class Node(object):
             self.log.warn('Disabling encryption!')
             self.server_io.setup_encryption(None)
 
+    def _set_task_dir(self, ctx) -> None:
+        """
+        Set the task dir
+
+        Parameters
+        ----------
+        ctx: DockerNodeContext or NodeContext
+            Context object containing settings
+        """
+        # If we're in a 'regular' context, we'll copy the dataset to our data
+        # dir and mount it in any algorithm container that's run; bind mounts
+        # on a folder will work just fine.
+        #
+        # If we're running in dockerized mode we *cannot* bind mount a folder,
+        # because the folder is in the container and not in the host. We'll
+        # have to use a docker volume instead. This means:
+        #  1. we need to know the name of the volume so we can pass it along
+        #  2. need to have this volume mounted so we can copy files to it.
+        #
+        #  Ad 1: We'll use a default name that can be overridden by an
+        #        environment variable.
+        #  Ad 2: We'll expect `ctx.data_dir` to point to the right place. This
+        #        is OK, since ctx will be a DockerNodeContext.
+        #
+        #  This also means that the volume will have to be created & mounted
+        #  *before* this node is started, so we won't do anything with it here.
+
+        # We'll create a subfolder in the data_dir. We need this subfolder so
+        # we can easily mount it in the algorithm containers; the root folder
+        # may contain the private key, which which we don't want to share.
+        # We'll only do this if we're running outside docker, otherwise we
+        # would create '/data' on the data volume.
+        if not ctx.running_in_docker:
+            self.__tasks_dir = ctx.data_dir / 'data'
+            os.makedirs(self.__tasks_dir, exist_ok=True)
+            self.__vpn_dir = ctx.data_dir / 'vpn'
+            os.makedirs(self.__vpn_dir, exist_ok=True)
+        else:
+            self.__tasks_dir = ctx.data_dir
+            self.__vpn_dir = ctx.vpn_dir
+
+    def setup_vpn_connection(
+            self, isolated_network_mgr: IsolatedNetworkManager, ctx
+    ) -> VPNManager:
+        """
+        Setup container which has a VPN connection
+
+        Returns
+        -------
+        VPNManager
+            Manages the VPN connection
+        """
+        ovpn_file = os.path.join(self.__vpn_dir, VPN_CONFIG_FILE)
+
+        self.log.debug("Setting up VPN client container")
+        vpn_volume_name = self.ctx.docker_vpn_volume_name \
+            if ctx.running_in_docker else self.__vpn_dir
+        vpn_manager = VPNManager(
+            isolated_network_mgr=isolated_network_mgr,
+            node_name=self.ctx.name,
+            vpn_volume_name=vpn_volume_name,
+            vpn_subnet=self.config.get('vpn_subnet')
+        )
+        # if vpn config doesn't exist, get it and write to disk
+        if not os.path.isfile(ovpn_file):
+            self._connect_vpn(vpn_manager, VPNConnectMode.REFRESH_COMPLETE,
+                              ovpn_file)
+        else:
+            self._connect_vpn(vpn_manager, VPNConnectMode.FIRST_TRY, ovpn_file)
+
+        return vpn_manager
+
+    def _connect_vpn(self, vpn_manager: VPNManager,
+                     connect_mode: VPNConnectMode, ovpn_file: str) -> None:
+        """
+        Connect to the VPN by starting up a VPN client container. If no VPN
+        config file exists, we only try once after first obtaining a config
+        file. If a VPN config file already exists, we first try to connect,
+        then try to refresh the keypair, and finally try to renew the entire
+        config file, until a connection is established.
+
+        Parameters
+        ----------
+        vpn_manager: VPNManager
+            Manages the VPN connection
+        connect_mode: VPNConnectMode
+            Specifies which parts of a config file to refresh before attempting
+            to connect
+        ovpn_file: str
+            Path to the VPN configuration file
+        """
+        do_try = True
+        if connect_mode == VPNConnectMode.FIRST_TRY:
+            self.log.debug("Using existing config file to connect to VPN")
+            next_mode = VPNConnectMode.REFRESH_KEYPAIR
+        elif connect_mode == VPNConnectMode.REFRESH_KEYPAIR:
+            self.log.debug("Refreshing VPN keypair...")
+            do_try = self.server_io.refresh_vpn_keypair(ovpn_file=ovpn_file)
+            next_mode = VPNConnectMode.REFRESH_COMPLETE
+        elif connect_mode == VPNConnectMode.REFRESH_COMPLETE:
+            self.log.debug("Requesting new VPN configuration file...")
+            do_try = self._get_vpn_config_file(ovpn_file)
+            next_mode = None  # if new config file doesn't work, give up
+
+        if do_try:
+            # try connecting to VPN
+            try:
+                vpn_manager.connect_vpn()
+            except Exception as e:
+                self.log.debug("Could not connect to VPN.")
+                self.log.debug(f"Exception: {e}")
+                # try again in another fashion
+                if next_mode:
+                    self._connect_vpn(vpn_manager, next_mode, ovpn_file)
+
+    def _get_vpn_config_file(self, ovpn_file: str) -> bool:
+        """
+        Obtain VPN configuration file from the server
+
+        Parameters
+        ----------
+        ovpn_file: str
+            Path to the VPN configuration file
+
+        Returns
+        -------
+        bool
+            Whether or not configuration file was successfully obtained
+        """
+        # get the ovpn configuration from the server
+        success, ovpn_config = self.server_io.get_vpn_config()
+        if not success:
+            self.log.warn("Obtaining VPN configuration file not successful!")
+            self.log.warn("Disabling node-to-node communication via VPN")
+            return False
+
+        # write ovpn config to node docker volume
+        with open(ovpn_file, 'w') as f:
+            f.write(ovpn_config)
+        return True
+
     def connect_to_socket(self):
         """ Create long-lasting websocket connection with the server.
 
@@ -545,23 +629,27 @@ class Node(object):
             new tasks.
         """
 
-        self.socketIO = SocketIO(request_timeout=60)
-
-        self.socketIO.register_namespace(NodeTaskNamespace('/tasks'))
-        NodeTaskNamespace.node_worker_ref = self
-
+        self.socketIO = SocketIO()
         self.socketIO.connect(
             url=f'{self.server_io.host}:{self.server_io.port}',
             headers=self.server_io.headers,
+            namespaces=['/tasks']
         )
 
+        NodeTaskNamespace.node_worker_ref = self
+        self.socketIO.register_namespace(NodeTaskNamespace('/tasks'))
+
         # Log the outcome
         while not self.socketIO.connected:
-            self.log.debug('Waiting for socket connection...')
+            self.log.debug('waiting for socket connection...')
             time.sleep(1)
 
-        self.log.info(f'Connected to host={self.server_io.host} on port='
-                      f'{self.server_io.port}')
+        msg = 'connected to host={host} on port={port}'
+        msg = msg.format(
+            host=self.server_io.host,
+            port=self.server_io.port
+        )
+        self.log.info(msg)
 
     def get_task_and_add_to_queue(self, task_id):
         """Fetches (open) task with task_id from the server.
@@ -614,9 +702,17 @@ class Node(object):
 
         except (KeyboardInterrupt, InterruptedError):
             self.log.info("Vnode is interrupted, shutting down...")
-            self.socketIO.disconnect()
+            self.cleanup()
             sys.exit()
 
+    def cleanup(self):
+        if hasattr(self, 'socketIO') and self.socketIO:
+            self.socketIO.disconnect()
+        if hasattr(self, 'vpn_manager') and self.vpn_manager:
+            self.vpn_manager.exit_vpn()
+        if hasattr(self, '_Node__docker') and self.__docker:
+            self.__docker.cleanup()
+
 
 # ------------------------------------------------------------------------------
 def run(ctx):
@@ -625,14 +721,8 @@ def run(ctx):
     logging.getLogger("requests").setLevel(logging.WARNING)
     logging.getLogger("engineio.client").setLevel(logging.WARNING)
 
-
     # initialize node, connect to the server using websockets
-    try:
-        node = Node(ctx)
-    except Exception as e:
-        log = logging.getLogger(logger_name(__name__))
-        log.critical(e)
-        exit(1)
+    node = Node(ctx)
 
     # put the node to work, executing tasks that are in the que
     node.run_forever()
